INITIAL_REVISION = 'initial_revision'
INITIAL_DATE = 'initial_date'
APPLIED_MIGRATIONS = 'applied_migrations'
COOKIECUTTER_TEMPLATE = 'cookiecutter_template'
COOKIECUTTER_DIRECTORY = 'cookiecutter_directory'
COOKIECUTTER_CONTEXT = 'cookiecutter_context'
